package com.jpattern.orm.session;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;

import com.jpattern.logger.ILogger;
import com.jpattern.orm.classtool.IOrmClassTool;
import com.jpattern.orm.classtool.IOrmClassToolMap;
import com.jpattern.orm.exception.OrmException;
import com.jpattern.orm.exception.OrmNotUniqueResultException;
import com.jpattern.orm.exception.OrmOptimisticLockException;
import com.jpattern.orm.generator.ResultSetMethodHelper;
import com.jpattern.orm.generator.wrapper.WrapperTypeArray;
import com.jpattern.orm.logger.OrmLogger;
import com.jpattern.orm.query.DeleteQuery;
import com.jpattern.orm.query.CustomFindQuery;
import com.jpattern.orm.query.FindQuery;
import com.jpattern.orm.query.SqlExecutor;
import com.jpattern.orm.query.UpdateQuery;
import com.jpattern.orm.query.OrmClassToolMapNameSolver;
import com.jpattern.orm.query.OrmCustomFindQuery;
import com.jpattern.orm.query.OrmDeleteQuery;
import com.jpattern.orm.query.OrmFindQuery;
import com.jpattern.orm.query.OrmUpdateQuery;
import com.jpattern.orm.query.PlainSqlExecutor;
import com.jpattern.orm.script.IScriptExecutor;
import com.jpattern.orm.script.ScriptExecutor;
import com.jpattern.orm.transaction.ITransaction;
import com.jpattern.orm.transaction.ITransactionDefinition;
import com.jpattern.orm.transaction.TransactionDefinition;

/**
 * 
 * @author Francesco Cina
 *
 * 27/giu/2011
 */
public class OrmSession implements SessionSqlPerformer {

	private final ILogger logger = OrmLogger.getOrmLogger(this.getClass());
	private final IOrmClassToolMap ormClassToolMap;
	private final SessionStrategy sessionStrategy;
	private final ResultSetMethodHelper rsMethodHelper;
	private final WrapperTypeArray wrapperTypeArray;

	public OrmSession(final IOrmClassToolMap ormClassToolMap, final SessionStrategy sessionStrategy, final ResultSetMethodHelper rsMethodHelper) {
		this.ormClassToolMap = ormClassToolMap;
		this.sessionStrategy = sessionStrategy;
		this.rsMethodHelper = rsMethodHelper;
		this.wrapperTypeArray = new WrapperTypeArray(rsMethodHelper);
	}

	@Override
	public final ITransaction transaction() throws OrmException {
		return this.transaction(new TransactionDefinition());
	}

	@Override
	public final <T> FindQuery<T> findQuery(final Class<T> clazz) throws OrmException {
		final OrmFindQuery<T> query = new OrmFindQuery<T>(this.ormClassToolMap , this, clazz, this.rsMethodHelper);
		final OrmClassToolMapNameSolver nameSolver = new OrmClassToolMapNameSolver(this.ormClassToolMap);
		nameSolver.register(clazz);
		query.setNameSolver(nameSolver);
		return query;
	}

	@Override
	public final <T> FindQuery<T> findQuery(final Class<T> clazz, final String alias) throws OrmException {
		final OrmFindQuery<T> query = new OrmFindQuery<T>(this.ormClassToolMap , this, clazz, this.rsMethodHelper);
		final OrmClassToolMapNameSolver nameSolver = new OrmClassToolMapNameSolver(this.ormClassToolMap);
		nameSolver.register(clazz, alias);
		query.setNameSolver(nameSolver);
		return query;
	}

	@Override
	public final CustomFindQuery findQuery(final String selectClause, final Class<?> clazz, final String alias ) throws OrmException {
		final OrmCustomFindQuery query = new OrmCustomFindQuery(selectClause, this.ormClassToolMap , this, clazz, this.rsMethodHelper);
		final OrmClassToolMapNameSolver nameSolver = new OrmClassToolMapNameSolver(this.ormClassToolMap);
		nameSolver.register(clazz, alias);
		query.setNameSolver(nameSolver);
		return query;
	}

	public final IOrmClassToolMap getOrmClassToolMap() {
		return this.ormClassToolMap;
	}

	@Override
	public final <T> T find(final Class<T> clazz, final Object value) throws OrmException {
		return this.find(clazz, new Object[]{value});
	}

	@Override
	public final <T> T find(final Class<T> clazz, final Object[] values) throws OrmException {
		final IOrmClassTool<T> ormClassTool = this.getOrmClassToolMap().getOrmClassTool(clazz);
		final ResultSetReader<T> resultSetReader = new ResultSetReader<T>() {

			@Override
			public T read(final ResultSet resultSet) throws SQLException {
				if ( resultSet.next() ) {
					final T result = ormClassTool.getOrmPersistor().mapRow("", resultSet, 0);
					if (resultSet.next()) {
						throw new OrmNotUniqueResultException("The query execution returned more than one object. Zero or one expected.");
					}
					return result;
				}
				return null;
			}
		};
		final SqlPerformer sqlExec = this.sqlPerformer();
		sqlExec.setMaxRows(1);
		return sqlExec.query(ormClassTool.getOrmCRUDQuery().getLoadQuery(), resultSetReader, this.wrapperTypeArray.unWrap(values));
	}

	@Override
	public final <T> T findUnique(final Class<T> clazz, final Object value) throws OrmException, OrmNotUniqueResultException {
		return this.findUnique(clazz, new Object[]{value});
	}

	@Override
	public final <T> T findUnique(final Class<T> clazz, final Object[] values) throws OrmException, OrmNotUniqueResultException {
		final T result = this.find(clazz, values);
		if (result==null) {
			throw new OrmNotUniqueResultException("No objects found.");
		}
		return result;
	}

	@Override
	public final <T> T save(final T object) throws OrmException {
		@SuppressWarnings("unchecked")
		final IOrmClassTool<T> ormClassTool = (IOrmClassTool<T>) this.getOrmClassToolMap().getOrmClassTool(object.getClass());
		final SqlPerformer sqlExec = this.sqlPerformer();
		final T resultObject = ormClassTool.getOrmPersistor().clone(object);
		//		final String sql = ormClassTool.getOrmCRUDQuery().getSaveQuery();

		//CHECK IF OBJECT HAS A 'VERSION' FIELD
		if (ormClassTool.getClassMapper().getTableMap().isVersionable()) {
			ormClassTool.getOrmPersistor().increaseVersion(resultObject, true);
		}


		if (!ormClassTool.getOrmPersistor().useKeyGenerators(resultObject)) {
			this.logger.debug("Saving object without using generators");
			final Object[] args = ormClassTool.getOrmPersistor().allValues(resultObject);
			final String sql = ormClassTool.getOrmCRUDQuery().getSaveQueryWithoutGenerators();
			sqlExec.update(sql, args);
		} else {
			this.logger.debug("Saving object using generators");
			final Object[] args = ormClassTool.getOrmPersistor().allNotGeneratedValues(resultObject);
			final String sql = ormClassTool.getOrmCRUDQuery().getSaveQuery();
			final GeneratedKeyReader generatedKeyExtractor = new GeneratedKeyReader() {

				@Override
				public void read(final ResultSet generatedKeyResultSet) throws SQLException {
					if (generatedKeyResultSet.next()) {
						ormClassTool.getOrmPersistor().updateGeneratedValues(generatedKeyResultSet, resultObject);
					}
				}

				@Override
				public String[] generatedColumnNames() {
					return ormClassTool.getClassMapper().getAllGeneratedColumnDBNames();
				}
			};
			sqlExec.update(sql, generatedKeyExtractor, args);
		}
		return resultObject;
	}

	@Override
	public final <T> T update(final T object) throws OrmException {
		return this.update(object, new AtomicBoolean());
	}

	private final <T> T update(final T object, final AtomicBoolean updated) throws OrmException {
		@SuppressWarnings("unchecked")
		final
		IOrmClassTool<T> ormClassTool = (IOrmClassTool<T>) this.getOrmClassToolMap().getOrmClassTool(object.getClass());
		final SqlPerformer sqlExec = this.sqlPerformer();
		final T resultObject = ormClassTool.getOrmPersistor().clone(object);

		//CHECK IF OBJECT HAS A 'VERSION' FIELD
		if (ormClassTool.getClassMapper().getTableMap().isVersionable()) {
			final int rightVersion = sqlExec.queryForInt(ormClassTool.getOrmCRUDQuery().getBeanVersionQuery(), ormClassTool.getOrmPersistor().primaryKeyAndVersionValues(resultObject));
			if ( rightVersion==0 ) {
				throw new OrmOptimisticLockException("The bean of class [" + resultObject.getClass() + "] cannot be updated. Version in the DB is not the expected one.");
			}
			ormClassTool.getOrmPersistor().increaseVersion(resultObject, false);
		}

		//UPDATE OBJECT
		final Object[] pkArgs = ormClassTool.getOrmPersistor().primaryKeyValues(resultObject);
		final Object[] npkArgs = ormClassTool.getOrmPersistor().notPrimaryKeyValues(resultObject);
		final Object[] args = new Object[ npkArgs.length + pkArgs.length ];
		int i=0;
		for (final Object value : npkArgs) {
			args[i++] = value;
		}
		for (final Object value : pkArgs) {
			args[i++] = value;
		}
		updated.set( sqlExec.update(ormClassTool.getOrmCRUDQuery().getUpdateQuery(), args) != 0);
		return resultObject;
	}


	@Override
	public final UpdateQuery updateQuery(final Class<?> clazz) throws OrmException {
		final OrmUpdateQuery update = new OrmUpdateQuery(clazz, this.ormClassToolMap, this, this.rsMethodHelper);
		final OrmClassToolMapNameSolver nameSolver = new OrmClassToolMapNameSolver(this.ormClassToolMap);
		nameSolver.alwaysResolveWithoutAlias(true);
		nameSolver.register(clazz);
		update.setNameSolver(nameSolver);
		return update;
	}

	@Override
	public final UpdateQuery updateQuery(final Class<?> clazz, final String alias) throws OrmException {
		final OrmUpdateQuery update = new OrmUpdateQuery(clazz, this.ormClassToolMap, this, this.rsMethodHelper);
		final OrmClassToolMapNameSolver nameSolver = new OrmClassToolMapNameSolver(this.ormClassToolMap);
		nameSolver.alwaysResolveWithoutAlias(true);
		nameSolver.register(clazz, alias);
		update.setNameSolver(nameSolver);
		return update;
	}

	@Override
	public final <T> int delete(final T object) throws OrmException {
		@SuppressWarnings("unchecked")
		final
		IOrmClassTool<T> ormClassTool = (IOrmClassTool<T>) this.getOrmClassToolMap().getOrmClassTool(object.getClass());
		final SqlPerformer sqlExec = this.sqlPerformer();
		return sqlExec.update(ormClassTool.getOrmCRUDQuery().getDeleteQuery(), ormClassTool.getOrmPersistor().primaryKeyValues(object));
	}

	@Override
	public final <T> int delete(final List<T> objects) throws OrmException {
		int result = 0;
		for (final T object : objects) {
			result += this.delete(object);
		}
		return result;
	}

	@Override
	public final DeleteQuery deleteQuery(final Class<?> clazz) throws OrmException {
		final OrmDeleteQuery delete = new OrmDeleteQuery(clazz, this.ormClassToolMap, this, this.rsMethodHelper);
		final OrmClassToolMapNameSolver nameSolver = new OrmClassToolMapNameSolver(this.ormClassToolMap);
		nameSolver.register(clazz);
		nameSolver.alwaysResolveWithoutAlias(true);
		delete.setNameSolver(nameSolver);
		return delete;
	}

	@Override
	public final DeleteQuery deleteQuery(final Class<?> clazz, final String alias) throws OrmException {
		final OrmDeleteQuery delete = new OrmDeleteQuery(clazz, this.ormClassToolMap, this, this.rsMethodHelper);
		final OrmClassToolMapNameSolver nameSolver = new OrmClassToolMapNameSolver(this.ormClassToolMap);
		nameSolver.register(clazz, alias);
		nameSolver.alwaysResolveWithoutAlias(true);
		delete.setNameSolver(nameSolver);
		return delete;
	}

	@Override
	public final IScriptExecutor scriptExecutor() throws OrmException {
		return new ScriptExecutor(this);
	}

	@Override
	public final SqlExecutor sqlExecutor() {
		return new PlainSqlExecutor(this);
	}

	@Override
	public ITransaction transaction(
			final ITransactionDefinition transactionDefinition) throws OrmException {
		return this.sessionStrategy.getTransaction(transactionDefinition);
	}

	@Override
	public SqlPerformer sqlPerformer() throws OrmException {
		return new PlainSqlPerformer(this.sessionStrategy.sqlPerformerStrategy());
	}

	@Override
	public <T> T saveOrUpdate(final T object) throws OrmException {
		@SuppressWarnings("unchecked")
		final IOrmClassTool<T> ormClassTool = (IOrmClassTool<T>) this.getOrmClassToolMap().getOrmClassTool(object.getClass());

		if (ormClassTool.getOrmPersistor().hasConditionalGenerator()) {
			if(ormClassTool.getOrmPersistor().useKeyGenerators(object)) {
				return this.save(object);
			} else {
				return this.update(object);
			}
		} else {
			try {
				return this.save(object);
			} catch (OrmException e) {
				return this.update(object);
			}
		}
	}

	@Override
	public <T> List<T> save(final Collection<T> objects) throws OrmException {
		final List<T> result = new ArrayList<T>();
		for (final T object : objects) {
			result.add(this.save(object));
		}
		return result;
	}

	@Override
	public <T> List<T> saveOrUpdate(final Collection<T> objects) throws OrmException {
		final List<T> result = new ArrayList<T>();
		for (final T object : objects) {
			result.add(this.saveOrUpdate(object));
		}
		return result;
	}

	@Override
	public <T> List<T> update(final Collection<T> objects) throws OrmException {
		final List<T> result = new ArrayList<T>();
		for (final T object : objects) {
			result.add(this.update(object));
		}
		return result;
	}

}