package com.jpattern.orm.session.datasource;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;

import com.jpattern.logger.ILogger;
import com.jpattern.orm.exception.OrmException;
import com.jpattern.orm.logger.OrmLogger;
import com.jpattern.orm.session.SqlPerformerStrategy;
import com.jpattern.orm.session.GeneratedKeyReader;
import com.jpattern.orm.session.PreparedStatementCreator;
import com.jpattern.orm.session.ResultSetReader;
import com.jpattern.orm.session.SqlSelectQuery;
import com.jpattern.orm.session.datasource.exception.SpringBasedSQLStateSQLExceptionTranslator;

/**
 * 
 * @author Francesco Cina
 *
 * 02/lug/2011
 * 
 * ISqlExecutor implementation using java.sql.Connection as backend.
 */
public class DataSourceSqlPerformerStrategy implements SqlPerformerStrategy, IConnectionCaller {

	private final ILogger logger = OrmLogger.getOrmLogger(this.getClass());
	private final DataSourceSessionProvider dataSourceSessionProvider;

	public DataSourceSqlPerformerStrategy(final DataSourceSessionProvider dataSourceSessionProvider) {
		this.dataSourceSessionProvider = dataSourceSessionProvider;
	}

	@Override
	public void execute(final String sql, final int timeout) throws OrmException {
		this.logger.debug("execute", "Method called");
		IPreparedStatement preparedStatement = null;
		IConnection conn = this.dataSourceSessionProvider.getConnection(false, this);
		try {
			preparedStatement = conn.prepareStatement( sql );
			preparedStatement.setQueryTimeout(timeout);
			preparedStatement.execute();
			conn.commit();
		} catch (SQLException e) {
			conn.rollback();
			throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("execute", sql, e);
		} catch (Exception e) {
			conn.rollback();
			throw new OrmException(e);
		} finally {
			try {
				if (preparedStatement!=null) {
					preparedStatement.close();
				}
			} catch (SQLException e) {
				throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("execute", sql, e);
			} finally {
				conn.close(this);
			}
		}
	}

	@Override
	public <T> T query(final SqlSelectQuery sqlSelectQuery, final ResultSetReader<T> rse) throws OrmException {
		this.logger.debug("query", "Method called");
		ResultSet resultSet = null;
		IPreparedStatement preparedStatement = null;
		IConnection conn = this.dataSourceSessionProvider.getConnection(true, this);
		String sql = "";
		try {
			sql = sqlSelectQuery.getSql();
			preparedStatement = conn.prepareStatement( sql );
			int i = 0;
			for (Object value : sqlSelectQuery.getArgs()) {
				preparedStatement.setObject(++i, value);
			}
			preparedStatement.setMaxRows(sqlSelectQuery.getMaxRows());
			preparedStatement.setQueryTimeout(sqlSelectQuery.getTimeout());
			resultSet = preparedStatement.executeQuery();
			return rse.read(resultSet);
		} catch (SQLException e) {
			throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("query", sql, e);
		} catch (Exception e) {
			throw new OrmException(e);
		} finally {
			try {
				if ((resultSet!=null) && !resultSet.isClosed()) {
					resultSet.close();
				}
				if (preparedStatement!=null) {
					preparedStatement.close();
				}
			} catch (SQLException e) {
				throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("query", sql, e);
			} finally {
				conn.close(this);
			}
		}
	}

	@Override
	public int update(final String sql, final int timeout, final Object... args) throws OrmException {
		this.logger.debug("update", "Method called");
		IConnection conn = this.dataSourceSessionProvider.getConnection(false, this);
		IPreparedStatement preparedStatement = null;
		try {
			preparedStatement = conn.prepareStatement( sql );
			preparedStatement.setQueryTimeout(timeout);
			int i = 0;
			for (Object value : args ) {
				preparedStatement.setObject(++i, value);
			}
			int result = preparedStatement.executeUpdate();
			conn.commit();
			return result;
		} catch (SQLException e) {
			conn.rollback();
			throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("update", sql, e);
		} catch (Exception e) {
			conn.rollback();
			throw new OrmException(e);
		} finally {
			try {
				if (preparedStatement!=null) {
					preparedStatement.close();
				}
			} catch (SQLException e) {
				throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("update", sql, e);
			} finally {
				conn.close(this);
			}
		}
	}

	@Override
	public int update(final String sql, final int timeout, final GeneratedKeyReader generatedKeyExtractor, final Object... args) throws OrmException {
		this.logger.debug("update", "Method called");
		IConnection conn = this.dataSourceSessionProvider.getConnection(false, this);
		ResultSet generatedKeyResultSet = null;
		IPreparedStatement preparedStatement = null;
		int result = 0;
		try {
			preparedStatement = conn.prepareStatement( sql , generatedKeyExtractor.generatedColumnNames());
			preparedStatement.setQueryTimeout(timeout);
			int i = 0;
			for (Object value : args) {
				preparedStatement.setObject(++i, value);
			}
			result = preparedStatement.executeUpdate();
			generatedKeyResultSet = preparedStatement.getGeneratedKeys();
			generatedKeyExtractor.read(generatedKeyResultSet);
			conn.commit();
			return result;
		} catch (SQLException e) {
			conn.rollback();
			throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("update", sql, e);
		} catch (Exception e) {
			conn.rollback();
			throw new OrmException(e);
		} finally {
			try {
				if (preparedStatement!=null) {
					preparedStatement.close();
				}
				if ((generatedKeyResultSet!=null) && !generatedKeyResultSet.isClosed()) {
					generatedKeyResultSet.close();
				}
			} catch (SQLException e) {
				throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("update", sql, e);
			} finally {
				conn.close(this);
			}
		}
	}

	@Override
	public int[] batchUpdate(final List<String> sqls, final int timeout) throws OrmException {
		this.logger.debug("batchUpdate", "Method called");
		IConnection conn = this.dataSourceSessionProvider.getConnection(false, this);
		IStatement statement = null;
		try {
			statement = conn.createStatement();
			statement.setQueryTimeout(timeout);
			for (String sql : sqls) {
				statement.addBatch(sql);
			}
			int[] result = statement.executeBatch();
			conn.commit();
			return result;
		} catch (SQLException e) {
			conn.rollback();
			throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("batchUpdate", "", e);
		} catch (Exception e) {
			conn.rollback();
			throw new OrmException(e);
		} finally {
			try {
				if (statement!=null) {
					statement.close();
				}
			} catch (SQLException e) {
				throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("batchUpdate", "", e);
			} finally {
				conn.close(this);
			}
		}
	}

	@Override
	public int[] batchUpdate(final String sql, final List<Object[]> args, final int timeout) throws OrmException {
		this.logger.debug("batchUpdate", "Method called");
		IConnection conn = this.dataSourceSessionProvider.getConnection(false, this);
		IPreparedStatement preparedStatement = null;
		try {
			preparedStatement = conn.prepareStatement( sql );
			preparedStatement.setQueryTimeout(timeout);
			for (Object[] arg : args) {
				int i = 0;
				for (Object value : arg) {
					preparedStatement.setObject(++i, value);
				}
				preparedStatement.addBatch();
			}
			int[] result = preparedStatement.executeBatch();
			conn.commit();
			return result;
		} catch (SQLException e) {
			conn.rollback();
			throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("batchUpdate", sql, e);
		} catch (Exception e) {
			conn.rollback();
			throw new OrmException(e);
		} finally {
			try {
				if (preparedStatement!=null) {
					preparedStatement.close();
				}
			} catch (SQLException e) {
				throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("batchUpdate", sql, e);
			} finally {
				conn.close(this);
			}
		}
	}

	@Override
	public int[] batchUpdate(final String sql, final PreparedStatementCreator psc, final int timeout) throws OrmException {
		this.logger.debug("batchUpdate", "Method called");
		IConnection conn = this.dataSourceSessionProvider.getConnection(false, this);
		PreparedStatementWrapper preparedStatement = null;
		try {
			preparedStatement = conn.prepareStatement( sql );
			preparedStatement.setQueryTimeout(timeout);
			for (int i=0; i<psc.getBatchSize(); i++) {
				psc.set(preparedStatement, i);
				preparedStatement.addBatch();
			}
			int[] result = preparedStatement.executeBatch();
			conn.commit();
			return result;
		} catch (SQLException e) {
			conn.rollback();
			throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("batchUpdate", sql, e);
		} catch (Exception e) {
			conn.rollback();
			throw new OrmException(e);
		} finally {
			try {
				if (preparedStatement!=null) {
					preparedStatement.close();
				}
			} catch (SQLException e) {
				throw SpringBasedSQLStateSQLExceptionTranslator.doTranslate("batchUpdate", sql, e);
			} finally {
				conn.close(this);
			}
		}
	}

}
